@@ -39,7 +39,7 @@
 -include_lib("gtp_utils.hrl").
 -include("conv.hrl").
 
--export([start_monitor/1, stop/1]).
+-export([start/1, stop/1]).
 -export([init/1,callback_mode/0,terminate/3]).
 -export([get_server_name_by_imsi/1, get_pid_by_imsi/1]).
 -export([auth_request/2, lu_request/1, tunnel_request/2, purge_ms_request/1]).
@@ -68,10 +68,10 @@ get_pid_by_imsi(Imsi) ->
         ServerName = get_server_name_by_imsi(Imsi),
         whereis(ServerName).
 
-start_monitor(Imsi) ->
+start(Imsi) ->
         ServerName = get_server_name_by_imsi(Imsi),
-        lager:info("ue_fsm start_monitor(~p)~n", [ServerName]),
-        gen_statem:start_monitor({local, ServerName}, ?MODULE, Imsi, [{debug, [trace]}]).
+        lager:info("ue_fsm start(~p)~n", [ServerName]),
+        gen_statem:start({local, ServerName}, ?MODULE, Imsi, [{debug, [trace]}]).
 
 stop(SrvRef) ->
         try
@@ -81,10 +81,10 @@ stop(SrvRef) ->
                 {error, Err}
         end.
 
-auth_request(Pid, {PdpTypeNr, Apn}) ->
+auth_request(Pid, {PdpTypeNr, Apn, EAP}) ->
         lager:info("ue_fsm auth_request~n", []),
         try
-                gen_statem:call(Pid, {auth_request, PdpTypeNr, Apn})
+                gen_statem:call(Pid, {auth_request, PdpTypeNr, Apn, EAP})
         catch
         exit:Err ->
                 {error, Err}
@@ -176,10 +176,10 @@ received_gtpc_delete_bearer_request(Pid) ->
 %% Internal helpers
 %% ------------------------------------------------------------------
 
-ev_handle({call, From}, {auth_request, PdpTypeNr, Apn}, Data) ->
-        case epdg_diameter_swm:auth_request(Data#ue_fsm_data.imsi, PdpTypeNr, Apn) of
+ev_handle({call, From}, {auth_request, PdpTypeNr, Apn, EAP}, Data) ->
+        case epdg_diameter_swm:auth_request(Data#ue_fsm_data.imsi, PdpTypeNr, Apn, EAP) of
         ok -> {next_state, state_wait_auth_resp, Data, [{reply,From,ok}]};
-        {error, Err} -> {stop_and_reply, Err, Data, [{reply,From,{error,Err}}]}
+        {error, Err} -> {stop_and_reply, Err, [{reply,From,{error,Err}}], Data}
 	end.
 
 %% ------------------------------------------------------------------
@@ -205,25 +205,27 @@ terminate(Reason, State, Data) ->
 state_new(enter, _OldState, Data) ->
         {keep_state, Data};
 
-state_new({call, _From} = EvType, {auth_request, PdpTypeNr, Apn} = EvContent, Data) ->
-        lager:info("ue_fsm state_new event=auth_request {~p, ~p}, ~p~n", [PdpTypeNr, Apn, Data]),
+state_new({call, _From} = EvType, {auth_request, PdpTypeNr, Apn, EAP} = EvContent, Data) ->
+        lager:info("ue_fsm state_new event=auth_request {~p, ~p, ~p}, ~p~n", [PdpTypeNr, Apn, EAP, Data]),
         ev_handle(EvType, EvContent, Data);
 
 state_new({call, From}, purge_ms_request, Data) ->
         lager:info("ue_fsm state_new event=purge_ms_request, ~p~n", [Data]),
-        {stop_and_reply, purge_ms_request, Data, [{reply,From,ok}]}.
+        {stop_and_reply, purge_ms_request, [{reply,From,ok}], Data}.
 
 state_wait_auth_resp(enter, _OldState, Data) ->
         {keep_state, Data};
 
 state_wait_auth_resp({call, From}, {received_swm_auth_response, Result}, Data) ->
         lager:info("ue_fsm state_wait_auth_resp event=received_swm_auth_response Result=~p, ~p~n", [Result, Data]),
-        gsup_server:auth_response(Data#ue_fsm_data.imsi, Result),
         case Result of
-                {ok, _} ->
+                {ok, _AuthTuples} ->
+                        gsup_server:auth_response(Data#ue_fsm_data.imsi, Result),
                         {next_state, state_authenticating, Data, [{reply,From,ok}]};
-                {error, Err} ->
-                        {next_state, state_new, Data, [{reply,From,{error,Err}}]};
+                {error, DiaRC} ->
+                        GsupCause = conv:dia_rc_to_gsup_cause(DiaRC),
+                        gsup_server:auth_response(Data#ue_fsm_data.imsi, {error, GsupCause}),
+                        {next_state, state_new, Data, [{reply,From,ok}]};
                 _ ->
                         {next_state, state_new, Data, [{reply,From,{error,unknown}}]}
         end.
@@ -231,8 +233,8 @@ state_wait_auth_resp({call, From}, {received_swm_auth_response, Result}, Data) -
 state_authenticating(enter, _OldState, Data) ->
         {keep_state, Data};
 
-state_authenticating({call, _From} = EvType, {auth_request, PdpTypeNr, Apn} = EvContent, Data) ->
-        lager:info("ue_fsm state_authenticating event=auth_request {~p, ~p}, ~p~n", [PdpTypeNr, Apn, Data]),
+state_authenticating({call, _From} = EvType, {auth_request, PdpTypeNr, Apn, EAP} = EvContent, Data) ->
+        lager:info("ue_fsm state_authenticating event=auth_request {~p, ~p, ~p}, ~p~n", [PdpTypeNr, Apn, EAP, Data]),
         ev_handle(EvType, EvContent, Data);
 
 state_authenticating({call, From}, lu_request, Data) ->
@@ -240,7 +242,7 @@ state_authenticating({call, From}, lu_request, Data) ->
         % Rx "GSUP CEAI LU Req" is our way of saying Rx "Swm Diameter-EAP REQ (DER) with EAP AVP containing successuful auth":
         case epdg_diameter_swm:auth_compl_request(Data#ue_fsm_data.imsi, Data#ue_fsm_data.apn) of
         ok -> {keep_state, Data, [{reply,From,ok}]};
-        {error, Err} -> {stop_and_reply, Err, Data, [{reply,From,{error,Err}}]}
+        {error, Err} -> {stop_and_reply, Err, [{reply,From,{error,Err}}], Data}
         end;
 
 % Rx Swm Diameter-EAP Answer (DEA) containing APN-Configuration, triggered by
@@ -257,20 +259,19 @@ state_authenticating({call, From}, {received_swm_auth_compl_response, Result}, D
                 PGWAddrCandidateList ->
                         Data1 = Data#ue_fsm_data{pgw_rem_addr_list = PGWAddrCandidateList}
                 end,
-                Ret = ok,
-                gsup_server:lu_response(Data1#ue_fsm_data.imsi, Ret),
-                {next_state, state_authenticated, Data1, [{reply,From,Ret}]};
-        {error, Err} ->
-                Ret = {error, Err},
-                gsup_server:lu_response(Data#ue_fsm_data.imsi, Ret),
-                {next_state, state_new, Data, [{reply,From,Ret}]}
+                gsup_server:lu_response(Data1#ue_fsm_data.imsi, ok),
+                {next_state, state_authenticated, Data1, [{reply,From,ok}]};
+        {error, DiaRC} ->
+                GsupCause = conv:dia_rc_to_gsup_cause(DiaRC),
+                gsup_server:lu_response(Data#ue_fsm_data.imsi, {error, GsupCause}),
+                {next_state, state_new, Data, [{reply,From,ok}]}
         end.
 
 state_authenticated(enter, _OldState, Data) ->
         {keep_state, Data};
 
-state_authenticated({call, _From}, {auth_request, PdpTypeNr, Apn}, Data) ->
-        lager:info("ue_fsm state_authenticated event=auth_request {~p, ~p}, ~p~n", [PdpTypeNr, Apn, Data]),
+state_authenticated({call, _From}, {auth_request, PdpTypeNr, Apn, EAP}, Data) ->
+        lager:info("ue_fsm state_authenticated event=auth_request {~p, ~p, ~p}, ~p~n", [PdpTypeNr, Apn, EAP, Data]),
         {next_state, state_new, Data, [postpone]};
 
 state_authenticated({call, From}, {tunnel_request, PCO}, Data) ->
@@ -337,8 +338,8 @@ state_wait_create_session_resp(state_timeout, create_session_timeout, Data) ->
 state_active(enter, _OldState, Data) ->
         {keep_state, Data};
 
-state_active({call, _From}, {auth_request, PdpTypeNr, Apn}, Data) ->
-        lager:info("ue_fsm state_active event=auth_request {~p, ~p}, ~p~n", [PdpTypeNr, Apn, Data]),
+state_active({call, _From}, {auth_request, PdpTypeNr, Apn, EAP}, Data) ->
+        lager:info("ue_fsm state_active event=auth_request {~p, ~p, ~p}, ~p~n", [PdpTypeNr, Apn, EAP, Data]),
         gtp_u_tun:delete_pdp_context(Data#ue_fsm_data.tun_pdp_ctx),
         Data1 = Data#ue_fsm_data{tun_pdp_ctx = undefined},
         {next_state, state_new, Data1, [postpone]};
@@ -401,11 +402,11 @@ state_wait_swm_session_termination_answer(enter, _OldState, Data) ->
                 {keep_state, Data}
         end;
 
-state_wait_swm_session_termination_answer({call, From}, {received_swm_sta, DiaResultCode}, Data) ->
+state_wait_swm_session_termination_answer({call, From}, {received_swm_sta, DiaRC}, Data) ->
         lager:info("ue_fsm state_wait_swm_session_termination_answer event=received_swm_sta, ~p~n", [Data]),
         case Data#ue_fsm_data.tear_down_gsup_needed of
         true ->
-                case {DiaResultCode, Data#ue_fsm_data.tear_down_gsup_cause} of
+                case {DiaRC#epdg_dia_rc.result_code, Data#ue_fsm_data.tear_down_gsup_cause} of
                 {2001, 0} -> gsup_server:purge_ms_response(Data#ue_fsm_data.imsi, ok);
                 {2001, _} -> gsup_server:purge_ms_response(Data#ue_fsm_data.imsi, {error, Data#ue_fsm_data.tear_down_gsup_cause});
                 _ -> gsup_server:purge_ms_response(Data#ue_fsm_data.imsi, {error, ?GSUP_CAUSE_NET_FAIL})
